#include <time.h>
#include <stdio.h>
#include <stdlib.h>


typedef unsigned long long ull;
// Смятаме по модул, защото иначе програмата избива от тип.
// В момента не ни е цел да предотвратяваме това.

const int MOD = 100000;

/**
    Най-обикновена рекурсия, която директно пренася дефиницията на редицата в код.
    Работи за O(2^n) време.
*/
int fib_1(int n)
{
    if(n == 0 || n == 1)
        return 1;
    return fib_1(n-1) + fib_1(n-2);
}

/**
    По-интелигентна версия, която запомня смятани преди резултати и не ги смята по много пъти.
    Работи за O(n) време.
*/
int fib_2(int n)
{
    if(n == 0 || n == 1)
        return 1;

    int a,b=1,c=1;
    n-=2;

    do
    {
        a = (b + c) % MOD;
        c = b;
        b = a;        
    }while(n--);

    return a;
}

/**
    Помощна структура, представлява матрица 2x2.
*/
struct ma3x
{
    ull pts[4];
};

/**
    Умножава две матрици.
    Резултатът се записва в първия аргумент.
*/
void multiply(struct ma3x *m1, struct ma3x *m2, int mod);

/**
    Много бърза имплементация, която се базира на факта, че матрицата:
    {
        { 1 1 }
        { 1 0 }
    }
    Повдигната на (N-2)-ра степен съдържа N-тото число на Фибоначи в първия елемент на първия си ред.
    Използва бързо повдигане на степен.

    Работи за O(log(n)) време.
*/
ull fib_3(ull n)
{
    n-=2;
    if(n < 0)
        return 1;
    struct ma3x res = {1,1,1,0};
    struct ma3x m1 = {1,1,1,0};
    while(n)
    {
        if(n % 2)
            multiply(&res, &m1, MOD);
        n/=2;
        multiply(&m1, &m1, MOD);
    }
    return res.pts[0];
}

void multiply(struct ma3x *m1, struct ma3x *m2, int mod)
{
    struct ma3x tmp = {0,0,0,0};
    #define a(n) m1->pts[n]
    #define b(n) m2->pts[n]

    tmp.pts[0] = (a(0)*b(0) + a(1)*b(2)) % mod;
    tmp.pts[1] = (a(0)*b(1) + a(1)*b(3)) % mod;
    tmp.pts[2] = (a(2)*b(0) + a(3)*b(2)) % mod;
    tmp.pts[3] = (a(2)*b(1) + a(3)*b(3)) % mod;

    int x;
    for(x=0;x<4;++x)
        m1->pts[x] = tmp.pts[x];

    #undef a
    #undef b
    return;
}

enum {SLOW, FAST, UBER_FAST};
/**
    Функция, която се използва за тестване на кода.
    От *implementation* зависи коя функция за смятане ще се викне.

    Числата на Фибоначи през *step* докато стигне до *n*.
    Примерно, при n = 100, step = 10 ще сметне:
     10, 20, 30, 40, 50, 60, 70, 80, 90, 100
*/
void test_fib(int implementation, ull n, ull step)
{
    struct timespec begin, end;
    ull x;
    for(x=step;x<=n;x+=step)
    {
        clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &begin);
        switch(implementation)
        {
            case SLOW:
                fib_1(x);
                break;
            case FAST:
                fib_2(x);
                break;
            case UBER_FAST:
                fib_3(x);
                break;
        }
        clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &end);
        printf("%lld %ld.%.9ld\n", x, end.tv_sec - begin.tv_sec, end.tv_nsec - begin.tv_nsec);
    }
}

int main(int argc, char ** argv)
{
    /*
        SLOW = 0
        FAST = 1
        UBER_FAST = 2
    */
    int type = UBER_FAST;
    ull n;
    ull step = 1;

    // Четене на аргументите. Нищо интересно.
    if(argc < 2)
        return 1;
    type = atoi(argv[1]);

    if(argc < 3)
        n = 32;
    else if(argc == 3)
        n = strtoull(argv[2], 0, 10);
    else
    {
        n = strtoull(argv[2], 0, 10);
        step = strtoull(argv[3], 0, 10);
    }

    // Самото смятане на числата.
    test_fib(type, n, step);
    return 0;
}
